#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/exceptions.py
# VERSION:     0.5.2
# CREATED:     2025-05-05 01:23
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************


class BuildUnsuccessful(OSError):
  def __init__(self, code: int, message: str):
    self.errno = code
    self.strerr = message


__all__: tuple[str, ...] = ("BuildUnsuccessful",)



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/views/yggdrasil.py
# VERSION:     0.5.2
# CREATED:     2024-06-26 00:14
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from collections import deque
from math import floor
from re import search
from textwrap import wrap
from typing import Deque, Generator, Optional, Union

### Third-party packages ###
from rich.box import MINIMAL
from rich.console import ConsoleRenderable, Group, RichCast
from rich.progress import BarColumn, Progress, Task, TaskID
from rich.table import Table


### Local modules ###
from aesir.exceptions import BuildUnsuccessful
from aesir.types import Chunk


class Yggdrasil(Progress):
  primary_task: Task
  rows: Deque[str]
  table: Table = Table(box=MINIMAL, show_lines=False, show_header=False)

  def __init__(self, row_count: int) -> None:
    self.rows = deque(maxlen=row_count)
    super().__init__()

  def get_renderable(self) -> Union[ConsoleRenderable, RichCast, str]:
    return Group(self.table, *self.get_renderables())

  def get_renderables(self) -> Generator[Table, None, None]:
    for task in self.tasks:
      if task.fields.get("progress_type") == "build":
        image_name: str = task.description or "undefined"
        if task.completed < 0:
          self.columns = (
            f"[red bold]Build unsuccessful for <Image '{image_name}'>.",
            "".ljust(9),
            BarColumn(),
          )
        elif task.completed == 0:
          self.columns = (
            f"Preparing to build <[bright_magenta]Image [green]'{image_name}'[reset]>…",
            "".ljust(9),
            BarColumn(),
          )
        elif task.completed > 0 and task.completed < 100:
          self.columns = (
            f"Building <[bright_magenta]Image [green]'{image_name}'[reset]>…",
            "".ljust(9),
            BarColumn(),
          )
        else:
          self.columns = (
            f"[blue]Built [reset]<[bright_magenta]Image [green]'{ image_name }'[reset]>"
            "[blue] successfully.[reset]",
            BarColumn(),
          )
      elif task.fields.get("progress_type") == "primary":
        self.columns = ("Build specified images:".ljust(42), BarColumn())
      yield self.make_tasks_table([task])

  def progress_build(self, chunks: Generator[dict[str, str], None, None], task_id: TaskID) -> None:
    """
    :raises BuildUnsuccessful:
    :raises pydantic.ValidationError:
    """
    for dictionary in chunks:
      chunk: Chunk = Chunk.model_validate(dictionary)
      if chunk.stream is not None:
        step = search(r"^Step (?P<divided>\d+)\/(?P<divisor>\d+) :", chunk.stream)
        if step is not None:
          divided: int = int(step.group("divided"))
          divisor: int = int(step.group("divisor"))
          self.update(task_id, completed=floor(divided / divisor * 100))
        self.update_table(chunk.stream)
      elif chunk.error is not None:
        self.update_table(f"[red]{chunk.error}[reset]")
        if chunk.error_detail is not None:
          raise BuildUnsuccessful(code=chunk.error_detail.code, message=chunk.error_detail.message)
        else:
          raise BuildUnsuccessful(code=255, message=chunk.error)

  def update_table(self, row: Optional[str] = None) -> None:
    if row is not None:
      self.rows.append(row)
    table: Table = Table(box=MINIMAL, show_lines=False, show_header=False)
    list(map(lambda row: table.add_row("\n".join(wrap(row, width=92)), style="grey50"), self.rows))
    self.table = table


__all__: tuple[str, ...] = ("Yggdrasil",)



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/views/bifrost.py
# VERSION:     0.5.2
# CREATED:     2024-06-25 19:43
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from re import match
from typing import ClassVar

### Third-party packages ###
from blessed import Terminal
from blessed.keyboard import Keystroke
from curses import KEY_DOWN, KEY_UP
from docker.models.containers import Container
from pydantic import BaseModel, ConfigDict, StrictInt, StrictStr, TypeAdapter
from rich.layout import Layout
from rich.live import Live
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

### Local modules ###
from aesir.types import BlockchainInfo, LNDInfo, MempoolInfo


class Bifrost(BaseModel):
  model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)
  bitcoind: Container
  container_index: StrictInt = 0
  container_names: list[StrictStr] = []
  containers: list[Container] = []

  ### Split layouts ###
  body: ClassVar[Layout] = Layout(name="body", minimum_size=4, ratio=8, size=17)
  footer: ClassVar[Layout] = Layout(name="footer", size=3)
  main: ClassVar[Layout] = Layout(size=72)
  pane: ClassVar[Layout] = Layout()
  realms: ClassVar[Layout] = Layout(name="realms", size=20)
  sidebar: ClassVar[Layout] = Layout(size=24)

  ### Terminal ###
  terminal: ClassVar[Terminal] = Terminal()

  def model_post_init(self, _) -> None:  # type: ignore[no-untyped-def]
    self.pane.split_row(self.sidebar, self.main)
    self.main.split_column(self.body, self.footer)
    self.sidebar.split_column(self.realms)

  def display(self) -> None:
    with (
      self.terminal.cbreak(),
      self.terminal.hidden_cursor(),
      Live(self.pane, refresh_per_second=4, transient=True),
    ):
      try:
        while True:
          ### Process input key ###
          keystroke: Keystroke = self.terminal.inkey(timeout=0.25)
          if keystroke.code == KEY_UP and self.container_index > 0:
            self.container_index -= 1
          elif keystroke.code == KEY_DOWN and self.container_index < len(self.container_names) - 1:
            self.container_index += 1
          elif keystroke in {"Q", "q"}:
            raise StopIteration

          container_rows: str = ""
          if self.container_index > 0:
            container_rows = "\n".join(self.container_names[: self.container_index])
            container_rows += f"\n[reverse]{self.container_names[self.container_index]}[reset]\n"
          else:
            container_rows = f"[reverse]{self.container_names[self.container_index]}[reset]\n"
          if self.container_index < len(self.container_names) - 1:
            container_rows += "\n".join(self.container_names[self.container_index + 1 :])
          self.pane["realms"].update(Panel(container_rows, title="realms"))

          container_name: str = self.container_names[self.container_index]
          body_table: Table = Table(expand=True, show_lines=True)
          body_table.add_column(container_name, "dark_sea_green bold")
          if match(r"aesir-(bitcoind)", container_name):
            blockchain_info: BlockchainInfo = TypeAdapter(BlockchainInfo).validate_json(
              self.bitcoind.exec_run(
                """
                bitcoin-cli -regtest -rpcuser=aesir -rpcpassword=aesir getblockchaininfo
                """
              ).output
            )
            mempool_info: MempoolInfo = TypeAdapter(MempoolInfo).validate_json(
              self.bitcoind.exec_run(
                """
                bitcoin-cli -regtest -rpcuser=aesir -rpcpassword=aesir getmempoolinfo
                """
              ).output
            )
            body_table.add_row(
              Text.assemble(
                f"\n{ 'Blockchain information:'.ljust(20) }\n",
                ("Chain: ", "bright_magenta bold"),
                blockchain_info.chain.ljust(9),
                ("Blocks: ", "green bold"),
                f"{blockchain_info.blocks}".ljust(8),
                ("Size: ", "blue bold"),
                f"{blockchain_info.size_on_disk}".ljust(10),
                ("Time: ", "cyan bold"),
                f"{blockchain_info.time}".rjust(10),
                "\n",
              )
            )
            body_table.add_row(
              Text.assemble(
                "\n",
                ("Mempool information".ljust(19), "bold"),
                "\n".ljust(19),
                ("Fees:".ljust(15), "green bold"),
                f"{mempool_info.total_fee}".rjust(15),
                "\n".ljust(19),
                ("Transactions:".ljust(15), "cyan bold"),
                f"{mempool_info.txn_count}".rjust(15),
                "\n".ljust(19),
                ("Size:".ljust(15), "blue bold"),
                f"{mempool_info.txn_bytes}".rjust(15),
                "\n".ljust(19),
                ("Loaded?:".ljust(15), "bright_magenta bold"),
                ("true".rjust(15), "green") if mempool_info.loaded else ("false".rjust(15), "red"),
                "\n".ljust(19),
                ("Usage:".ljust(15), "light_coral bold"),
                f"{mempool_info.usage}".rjust(15),
                "\n",
              )
            )
          elif match(r"aesir-(lnd|ping|pong)", container_name):
            container: Container = next(
              filter(lambda container: container.name == container_name, self.containers)
            )
            lnd_info: LNDInfo = TypeAdapter(LNDInfo).validate_json(
              container.exec_run(
                """
                lncli
                  --macaroonpath=/root/.lnd/data/chain/bitcoin/regtest/admin.macaroon
                  --rpcserver=localhost:10001
                  --tlscertpath=/root/.lnd/tls.cert
                getinfo
                """
              ).output
            )
            body_table.add_row(
              Text.assemble(
                "\n\n\n",
                ("Nodekey:\n", "light_coral bold"),
                lnd_info.identity_pubkey,
                "\n",
                "\n".ljust(19),
                ("Channels:".ljust(10), "green bold"),
                str(lnd_info.num_active_channels).rjust(20),
                "\n".ljust(19),
                ("Peers:".ljust(10), "cyan bold"),
                str(lnd_info.num_peers).rjust(20),
                "\n".ljust(19),
                ("Blocks:".ljust(10), "rosy_brown bold"),
                str(lnd_info.block_height).rjust(20),
                "\n".ljust(19),
                ("Synced?:".ljust(10), "steel_blue bold"),
                ("true".rjust(20), "green")
                if lnd_info.synced_to_chain
                else ("false".rjust(20), "red"),
                "\n\n\n",
              )
            )
          self.pane["body"].update(body_table)
          self.pane["footer"].update(
            Panel(
              Text.assemble(
                "Select:".rjust(16),
                (" ↑↓ ", "bright_magenta bold"),
                " " * 20,
                "Exit:".rjust(16),
                ("  Q ", "red bold"),
              )
            )
          )
      except StopIteration:
        print("Valhalla!")


__all__: tuple[str, ...] = ("Bifrost",)



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/types/image.py
# VERSION:     0.5.2
# CREATED:     2023-12-01 05:31
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Literal, Union

### Local modules ###
from aesir.types.build import BuildEnum

Image = Union[BuildEnum, Literal["postgres:latest", "redis:latest"]]

__all__: tuple[str, ...] = ("Image",)



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/commands/invoice.py
# VERSION:     0.5.2
# CREATED:     2024-11-15 00:56
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Dict, Tuple
from uuid import uuid4 as uuid

### Third-party packages ###
from click import argument, command, option
from docker import DockerClient, from_env
from docker.errors import APIError, NotFound
from docker.models.containers import Container
from pydantic import TypeAdapter
from rich import print as rich_print

### Local modules ###
from aesir.types import LNDInvoice, MutexOption, ServiceName


@argument("amount", default=1_000)
@argument("memo", default=str(uuid()))
@command
@option("--lnd", alternatives=["ping", "pong"], cls=MutexOption, is_flag=True, type=bool)
@option("--ping", alternatives=["lnd", "pong"], cls=MutexOption, is_flag=True, type=bool)
@option("--pong", alternatives=["lnd", "ping"], cls=MutexOption, is_flag=True, type=bool)
def invoice(amount: int, lnd: bool, memo: str, ping: bool, pong: bool) -> None:
  """For either "uno" or "duo" cluster, create an invoice from specified lnd container"""
  try:
    client: DockerClient = from_env()
    client.ping()
  except APIError:
    rich_print("[red bold]Unable to connect to docker daemon.")
    return

  ### Select LND container from specified mutually exclusive options ###
  container_name: ServiceName = "aesir-lnd"
  container_selector: Dict[ServiceName, bool] = {
    "aesir-lnd": lnd,
    "aesir-ping": ping,
    "aesir-pong": pong,
  }
  try:
    container_name = next(filter(lambda key_values: key_values[1], container_selector.items()))[0]
  except StopIteration:
    pass

  ### Initiate parameters ###
  container: Container
  try:
    container = client.containers.get(container_name)
  except NotFound:
    rich_print(f'[red bold]Unable to find specified LND container (name="{ container_name }")')
    return

  ### Generate invoice ###
  lnd_invoice: LNDInvoice = TypeAdapter(LNDInvoice).validate_json(
    container.exec_run(
      f"""
      lncli
        --macaroonpath=/home/lnd/.lnd/data/chain/bitcoin/regtest/admin.macaroon
        --rpcserver=localhost:10001
        --tlscertpath=/home/lnd/.lnd/tls.cert
      addinvoice
        --amt={ amount }
        --memo={ memo }
      """
    ).output
  )
  rich_print(lnd_invoice)


__all__: Tuple[str, ...] = ("invoice",)



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/configs.py
# VERSION:     0.5.2
# CREATED:     2023-12-01 05:31
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from pathlib import Path
from typing import Any, Optional

### Standard packages ###
from pydantic import TypeAdapter
from yaml import Loader, load

### Local modules ###
from aesir.types import (
  Build,
  BuildEnum,
  ClusterEnum,
  Service,
  ServiceName,
)

### Parse schemas ###
BUILDS: dict[BuildEnum, Build]
CLUSTERS: dict[ClusterEnum, dict[ServiceName, Service]]
NETWORK: str
PERIPHERALS: dict[ServiceName, Service]

file_path: Path = Path(__file__).resolve()
with open(str(file_path).replace("configs.py", "schemas.yml"), "rb") as stream:
  schema: Optional[dict[str, Any]] = load(stream, Loader=Loader)
  if schema:
    BUILDS = TypeAdapter(dict[BuildEnum, Build]).validate_python(schema["builds"])
    CLUSTERS = TypeAdapter(dict[ClusterEnum, dict[ServiceName, Service]]).validate_python(
      schema["clusters"]
    )
    NETWORK = schema.get("network", "aesir")
    PERIPHERALS = TypeAdapter(dict[ServiceName, Service]).validate_python(schema["peripherals"])

__all__: tuple[str, ...] = ("BUILDS", "CLUSTERS", "NETWORK", "PERIPHERALS")



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/types/cluster_enum.py
# VERSION:     0.5.2
# CREATED:     2023-12-01 05:31
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Literal

ClusterEnum = Literal["cat", "duo", "ohm", "uno"]

__all__: tuple[str, ...] = ("ClusterEnum",)



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/commands/deploy.py
# VERSION:     0.5.2
# CREATED:     2023-12-01 05:31
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from io import BytesIO
from re import match
from time import sleep
from typing import Iterator

### Third-party packages ###
from click import command, option
from docker import DockerClient, from_env
from docker.errors import APIError, BuildError, ImageNotFound, NotFound
from docker.models.containers import Container
from pydantic import TypeAdapter
from rich import print as rich_print
from rich.progress import TaskID, track

### Local modules ###
from aesir.configs import BUILDS, CLUSTERS, NETWORK, PERIPHERALS
from aesir.types import (
  Build,
  BuildEnum,
  ClusterEnum,
  MutexOption,
  NewAddress,
  Service,
  ServiceName,
)
from aesir.views import Yggdrasil


@command
@option("--cat", alternatives=["duo", "ohm", "uno"], cls=MutexOption, is_flag=True, type=bool)
@option("--duo", alternatives=["cat", "ohm", "uno"], cls=MutexOption, is_flag=True, type=bool)
@option("--ohm", alternatives=["cat", "duo", "uno"], cls=MutexOption, is_flag=True, type=bool)
@option("--uno", alternatives=["cat", "duo", "ohm"], cls=MutexOption, is_flag=True, type=bool)
@option("--with-cashu-mint", is_flag=True, help="Deploy cashu-mint peripheral service", type=bool)
@option("--with-electrs", is_flag=True, help="Deploy electrs peripheral service", type=bool)
@option("--with-ord-server", is_flag=True, help="Deploy ord-server peripheral service", type=bool)
@option("--with-postgres", is_flag=True, help="Deploy postgres peripheral service", type=bool)
@option("--with-redis", is_flag=True, help="Deploy redis peripheral service", type=bool)
def deploy(
  cat: bool,
  duo: bool,
  ohm: bool,
  uno: bool,
  with_cashu_mint: bool,
  with_electrs: bool,
  with_ord_server: bool,
  with_postgres: bool,
  with_redis: bool,
) -> None:
  """Deploy cluster, either with one or two LND nodes."""
  try:
    client: DockerClient = from_env()
    client.ping()
  except APIError:
    rich_print("[red bold]Unable to connect to docker daemon.")
    return

  ### Defaults to duo network; Derive cluster information from parameters ###
  cluster_selector: dict[ClusterEnum, bool] = {"cat": cat, "duo": duo, "ohm": ohm, "uno": uno}
  cluster_name: ClusterEnum = "duo"
  try:
    cluster_name = next(filter(lambda value: value[1], cluster_selector.items()))[0]
  except StopIteration:
    pass
  cluster: dict[ServiceName, Service] = CLUSTERS[cluster_name]
  image_selector: dict[ServiceName, bool] = {
    "aesir-cashu-mint": False,
    "aesir-electrs": with_electrs,
    "aesir-ord-server": False,
    "aesir-postgres": with_postgres,
    "aesir-redis": with_redis,
  }
  peripherals: Iterator[tuple[ServiceName, Service]] = filter(
    lambda peripheral_tuple: image_selector[peripheral_tuple[0]], PERIPHERALS.items()
  )
  cluster.update(peripherals)

  ### Attempts to create network if not exist ###
  try:
    client.networks.create(NETWORK, check_duplicate=True)
  except APIError:
    pass

  ### Deploy specified cluster ###
  for name, service in track(cluster.items(), f"Deploy {cluster_name} cluster:".ljust(42)):
    image_name: str = service.image
    flags: list[str] = list(service.command.values())
    ports: dict[str, str] = dict(
      map(lambda item: (item[0], item[1]), [port.split(":") for port in service.ports])
    )
    client.containers.run(
      image_name,
      command=flags,
      detach=True,
      environment=service.env_vars,
      name=name,
      network=NETWORK,
      ports=ports,
    )

  treasuries: list[str] = []
  if duo or uno:
    ### Wait until lnd(s) ready ###
    sleep(3)

    ### Mine starting capital ###
    for container in track(client.containers.list(), "Generate addresses:".ljust(42)):
      if match(r"aesir-lnd|aesir-ping|aesir-pong", container.name) is not None:
        new_address: NewAddress = TypeAdapter(NewAddress).validate_json(
          container.exec_run(
            """
            lncli
              --macaroonpath=/root/.lnd/data/chain/bitcoin/regtest/admin.macaroon
              --rpcserver=localhost:10001
              --tlscertpath=/root/.lnd/tls.cert
            newaddress p2wkh
            """
          ).output
        )
        treasuries.append(new_address.address)

  ### Define selection for shared-volume peripherals ###
  build_selector: dict[BuildEnum, bool] = {
    "aesir-bitcoind": False,
    "aesir-bitcoind-cat": False,
    "aesir-cashu-mint": with_cashu_mint,
    "aesir-electrs": with_electrs,
    "aesir-lnd": False,
    "aesir-ord-server": with_ord_server,
  }

  ### Build missing images if any for shared-volume peripherals ###
  image_names: list[str] = list(
    map(
      lambda image: image.tags[0].split(":")[0],
      filter(lambda image: len(image.tags) != 0, client.images.list()),
    )
  )
  builds: dict[str, Build] = {
    tag: build for tag, build in BUILDS.items() if build_selector[tag] and tag not in image_names
  }
  build_count: int = len(builds.keys())
  if build_count != 0:
    builds_items = builds.items()
    with Yggdrasil(row_count=10) as yggdrasil:
      task_id: TaskID = yggdrasil.add_task("", progress_type="primary", total=build_count)
      for tag, build in builds_items:
        build_task_id: TaskID = yggdrasil.add_task(tag, progress_type="build", total=100)
        with BytesIO("\n".join(build.instructions).encode("utf-8")) as fileobj:
          try:
            yggdrasil.progress_build(
              client.api.build(
                decode=True, fileobj=fileobj, platform=build.platform, rm=True, tag=tag
              ),
              build_task_id,
            )
          except BuildError:
            yggdrasil.update(
              build_task_id,
              completed=0,
              description=f"[red bold]Build unsuccessful for <Image '{tag}'>.",
            )
          yggdrasil.update(
            build_task_id,
            completed=100,
            description=f"[blue]Built <[bright_magenta]Image [green]'{tag}'[reset]> successfully.",
          )
          yggdrasil.update(task_id, advance=1)
      yggdrasil.update(task_id, completed=build_count, description="[blue]Complete")

  ### Deploy shared volume peripherals ###
  run_errors: list[str] = []
  volume_target: str = "aesir-ping" if duo else "aesir-lnd"
  for name, service in track(peripherals, "Deploy shared-volume peripherals:".ljust(42)):
    ports = dict(map(lambda item: (item[0], item[1]), [port.split(":") for port in service.ports]))
    volume_target = (
      "aesir-bitcoind" if name in {"aesir-electrs", "aesir-ord-server"} else volume_target
    )
    try:
      client.containers.run(
        service.image,
        command=service.command.values(),
        detach=True,
        environment=service.env_vars,
        name=name,
        network=NETWORK,
        ports=ports,
        volumes_from=[volume_target],
      )
    except ImageNotFound:
      run_errors.append(
        f"<[bright_magenta]Image [green]'{service.image}'[reset]> [red]is not found.[reset]"
      )
  list(map(rich_print, run_errors))

  if len(treasuries) != 0:
    ### Retrieve bitcoind container ###
    bitcoind: Container
    try:
      bitcoind = client.containers.get("aesir-bitcoind")
    except NotFound:
      rich_print('[dim yellow1]Unable to find "aesir-bitcoind"; initial capital not yet mined.')
      return
    for address in track(treasuries, "Mine initial capital for parties:".ljust(42)):
      bitcoind.exec_run(
        """
        bitcoin-cli -regtest -rpcuser=aesir -rpcpassword=aesir generatetoaddress 101 %s
        """
        % address
      )


__all__: tuple[str, ...] = ("deploy",)



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/types/chunk.py
# VERSION:     0.5.2
# CREATED:     2025-05-05 01:23
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Optional

### Third-party packages ###
from pydantic import BaseModel, Field, StrictInt, StrictStr


class ErrorDetail(BaseModel):
  code: StrictInt
  message: StrictStr


class Chunk(BaseModel):
  error: Optional[StrictStr] = None
  error_detail: Optional[ErrorDetail] = Field(alias="errorDetail", default=None)
  stream: Optional[StrictStr] = None


__all__: tuple[str, ...] = ("Chunk", "ErrorDetail")



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/__init__.py
# VERSION:     0.5.2
# CREATED:     2023-12-01 05:31
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION: https://www.w3docs.com/snippets/python/what-is-init-py-for.html
#
# HISTORY:
# *************************************************************

### Local modules ###
from aesir.commands import build, clean, deploy, invoice, mine, nodekeys, ping_pong, pull

__all__: tuple[str, ...] = (
  "build",
  "clean",
  "deploy",
  "invoice",
  "mine",
  "nodekeys",
  "ping_pong",
  "pull",
)
__version__: str = "0.5.2"



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/views/__init__.py
# VERSION:     0.5.2
# CREATED:     2024-06-25 19:43
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION: https://www.w3docs.com/snippets/python/what-is-init-py-for.html
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Tuple

### Local modules ###
from aesir.views.bifrost import Bifrost
from aesir.views.yggdrasil import Yggdrasil

__all__: Tuple[str, ...] = ("Bifrost", "Yggdrasil")



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/commands/build.py
# VERSION:     0.5.2
# CREATED:     2024-02-27 23:52
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from io import BytesIO

### Third-party packages ###
from click import command, option
from docker import DockerClient, from_env
from docker.errors import APIError, BuildError
from pydantic import ValidationError
from rich import print as rich_print
from rich.progress import TaskID

### Local modules ###
from aesir.configs import BUILDS
from aesir.exceptions import BuildUnsuccessful
from aesir.types import Build, BuildEnum
from aesir.views import Yggdrasil


@command
@option("--bitcoind", is_flag=True, help="Build bitcoind image", type=bool)
@option("--bitcoind-cat", is_flag=True, help="Build bitcoind-cat optional image", type=bool)
@option("--cashu-mint", is_flag=True, help="Build cashu-mint optional image", type=bool)
@option("--electrs", is_flag=True, help="Build electrs optional image", type=bool)
@option("--lnd", is_flag=True, help="Build lnd image", type=bool)
@option("--ord-server", is_flag=True, help="Build ord-server optional image", type=bool)
def build(
  bitcoind: bool,
  bitcoind_cat: bool,
  cashu_mint: bool,
  electrs: bool,
  lnd: bool,
  ord_server: bool,
) -> None:
  """Build peripheral images for the desired cluster."""
  try:
    client: DockerClient = from_env()
    client.ping()
  except APIError:
    rich_print("[red bold]Unable to connect to docker daemon.")
    return

  ### Build optional images ###
  image_names: list[str] = list(
    map(
      lambda image: image.tags[0].split(":")[0],
      filter(lambda image: len(image.tags) != 0, client.images.list()),
    )
  )
  build_select: dict[BuildEnum, bool] = {
    "aesir-bitcoind": bitcoind,
    "aesir-bitcoind-cat": bitcoind_cat,
    "aesir-cashu-mint": cashu_mint,
    "aesir-electrs": electrs,
    "aesir-lnd": lnd,
    "aesir-ord-server": ord_server,
  }

  outputs: list[str] = []
  built: set[str] = {tag for tag in BUILDS.keys() if build_select[tag] and tag in image_names}
  outputs += map(lambda tag: f"<Image: '{tag}'> already exists within images.", built)
  list(map(rich_print, outputs))

  builds: dict[str, Build] = {
    tag: build for tag, build in BUILDS.items() if build_select[tag] and tag not in image_names
  }
  build_count: int = len(builds.keys())
  if build_count != 0:
    builds_items = builds.items()
    with Yggdrasil(row_count=10) as yggdrasil:
      task_id: TaskID = yggdrasil.add_task("", progress_type="primary", total=build_count)
      for tag, build in builds_items:
        build_task_id: TaskID = yggdrasil.add_task(tag, progress_type="build", total=100)
        with BytesIO("\n".join(build.instructions).encode("utf-8")) as fileobj:
          try:
            yggdrasil.progress_build(
              client.api.build(
                decode=True, fileobj=fileobj, gzip=True, platform=build.platform, rm=True, tag=tag
              ),
              build_task_id,
            )
          except (BuildError, BuildUnsuccessful, ValidationError):
            yggdrasil.update(build_task_id, completed=-1)
            continue
          yggdrasil.update(build_task_id, completed=100)
          yggdrasil.update(task_id, advance=1)
      yggdrasil.update(task_id, completed=build_count, description="[blue]Complete[reset]")


__all__: tuple[str, ...] = ("build",)



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/commands/pull.py
# VERSION:     0.5.2
# CREATED:     2023-12-01 06:18
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Dict, Iterator, List, Set, Tuple

### Third-party packages ###
from click import command, option
from docker import DockerClient, from_env
from docker.errors import APIError
from rich import print as rich_print
from rich.progress import track

### Local modules ###
from aesir.configs import BUILDS, PERIPHERALS
from aesir.types import Image, Service, ServiceName


@command
@option("--postgres", is_flag=True, help="Pull postgres optional image", type=bool)
@option("--redis", is_flag=True, help="Pull redis optional image", type=bool)
def pull(postgres: bool, redis: bool) -> None:
  """Download required and flagged optional docker images from hub."""
  try:
    client: DockerClient = from_env()
    client.ping()
  except APIError:
    rich_print("[red bold]Unable to connect to docker daemon.")
    return

  ### Pull required images ###
  outputs: List[str] = []
  docker_images: Set[str] = {image.tags[0] for image in client.images.list()}
  for registry_id in track(BUILDS.keys(), "Pull buildable images:".ljust(42)):
    if registry_id in docker_images:
      outputs.append(
        f"<[bright_magenta]Image: [green]'{ registry_id }'[reset]> already exists in registry."
      )
    else:
      repository, tag = registry_id.split(":")
      client.images.pull(repository=repository, tag=tag)
      outputs.append(f"<[bright_magenta]Image: [green]'{ registry_id }'[reset]> downloaded.")
  list(map(rich_print, outputs))

  ### Pull peripheral images ###
  outputs = []
  peripheral_selector: Dict[ServiceName, bool] = {
    "aesir-cashu-mint": False,
    "aesir-ord-server": False,
    "aesir-postgres": postgres,
    "aesir-redis": redis,
  }
  peripherals: Iterator[Tuple[ServiceName, Service]] = filter(
    lambda service_tuple: peripheral_selector[service_tuple[0]],
    PERIPHERALS.items(),
  )
  for _, peripheral in track(peripherals, "Pull peripherals images flagged:".ljust(42)):
    registry_id: Image = peripheral.image  # type: ignore[no-redef]
    if registry_id in docker_images:
      outputs.append(
        f"<[bright_magenta]Image: [green]'{ registry_id }'[reset]> already exists in registry."
      )
    else:
      repository, tag = registry_id.split(":")
      client.images.pull(repository=repository, tag=tag)
      outputs.append(f"<[bright_magenta]Image: [green]'{ registry_id }'[reset]> downloaded.")
  list(map(rich_print, outputs))


__all__: Tuple[str, ...] = ("pull",)



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/types/build_enum.py
# VERSION:     0.5.2
# CREATED:     2023-12-06 05:31
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from typing import Literal

### Third-party packages ###
from pydantic import BaseModel, StrictStr


class Build(BaseModel):
  instructions: list[StrictStr]
  platform: StrictStr = "linux/amd64"


BuildEnum = Literal[
  "aesir-bitcoind",
  "aesir-bitcoind-cat",
  "aesir-cashu-mint",
  "aesir-electrs",
  "aesir-lnd",
  "aesir-ord-server",
]


__all__: tuple[str, ...] = ("Build", "BuildEnum")



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/types/lnd_invoice.py
# VERSION:     0.5.2
# CREATED:     2024-11-15 00:56
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Third-party packages ###
from pydantic import BaseModel, Field, StrictStr


class LNDInvoice(BaseModel):
  add_index: int
  r_hash: StrictStr
  payment_address: StrictStr = Field(alias="payment_addr")
  payment_request: StrictStr


__all__: tuple[str, ...] = ("LNDInvoice",)



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/commands/mine.py
# VERSION:     0.5.2
# CREATED:     2023-12-01 05:31
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Standard packages ###
from re import match
from typing import List, Tuple

### Third-party packages ###
from apscheduler.schedulers.background import BackgroundScheduler
from click import argument, command
from docker import DockerClient, from_env
from docker.errors import APIError, NotFound
from docker.models.containers import Container
from pydantic import TypeAdapter
from rich import print as rich_print
from rich.progress import track

### Local modules ###
from aesir.views import Bifrost
from aesir.types import NewAddress


@command
@argument("blockcount", default=1, type=int)
@argument("blocktime", default=5, type=int)
def mine(blockcount: int, blocktime: int) -> None:
  """Scheduled mining with "blockcount" and "blocktime"."""
  try:
    client: DockerClient = from_env()
    client.ping()
  except APIError:
    rich_print("[red bold]Unable to connect to daemon.")
    return

  ### Retrieve bitcoind container ###
  bitcoind: Container
  try:
    bitcoind = client.containers.get("aesir-bitcoind")
  except NotFound:
    try:
      bitcoind = client.containers.get("aesir-bitcoind-cat")
    except NotFound:
      rich_print('[red bold]Unable to find "aesir-bitcoind" container.')
      return

  ### Retrieve other containers ###
  aesir_containers: List[Container] = list(
    filter(lambda container: match(r"aesir-*", container.name), reversed(client.containers.list()))
  )
  container_names: List[str] = list(map(lambda container: container.name, aesir_containers))
  lnd_containers: List[Container] = list(
    filter(lambda container: match(r"aesir-(lnd|ping|pong)", container.name), aesir_containers)
  )

  ### Generate treasury addresses as mining destinations ###
  treasuries: List[str] = []
  if len(lnd_containers) == 0:  # FIXME: current bitcoind does not have BerkeleyDB
    bitcoind.exec_run(
      """
      bitcoin-cli -regtest -rpcuser=aesir -rpcpassword=aesir createwallet default
      """
    )
    treasury_address: str = bitcoind.exec_run(
      """
      bitcoin-cli -regtest -rpcuser=aesir -rpcpassword=aesir getnewaddress treasury bech32
      """
    ).output.decode("utf-8")
    treasuries.append(treasury_address)
  else:
    for container in track(lnd_containers, "Generate mining treasuries:".ljust(42)):
      new_address: NewAddress = TypeAdapter(NewAddress).validate_json(
        container.exec_run(
          """
          lncli
            --macaroonpath=/root/.lnd/data/chain/bitcoin/regtest/admin.macaroon
            --rpcserver=localhost:10001
            --tlscertpath=/root/.lnd/tls.cert
          newaddress p2wkh
          """
        ).output
      )
      treasuries.append(new_address.address)

  ### Set up mining schedule using command arguments ###
  scheduler: BackgroundScheduler = BackgroundScheduler()
  for address in treasuries:
    scheduler.add_job(
      bitcoind.exec_run,
      "interval",
      [
        """
        bitcoin-cli -regtest -rpcuser=aesir -rpcpassword=aesir generatetoaddress %d %s
        """
        % (blockcount, address)
      ],
      seconds=blocktime,
    )
  scheduler.start()

  bifrost: Bifrost = Bifrost(
    bitcoind=bitcoind,
    containers=aesir_containers,
    container_index=0,
    container_names=container_names,
  )
  bifrost.display()


__all__: Tuple[str, ...] = ("mine",)



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/types/blockchain_info.py
# VERSION:     0.5.2
# CREATED:     2023-12-01 05:31
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Third-party packages ###
from pydantic import BaseModel, StrictInt, StrictStr


class BlockchainInfo(BaseModel):
  blocks: StrictInt = 0
  chain: StrictStr = ""
  size_on_disk: StrictInt = 0
  time: StrictInt = 0


__all__: tuple[str, ...] = ("BlockchainInfo",)



#!/usr/bin/env python3.10
# coding:utf-8
# Copyright (C) 2022-2025 All rights reserved.
# FILENAME:    ~~/src/aesir/types/open_channel.py
# VERSION:     0.5.2
# CREATED:     2023-12-02 20:50
# AUTHOR:      Sitt Guruvanich <aekasitt.g+github@siamintech.co.th>
# DESCRIPTION:
#
# HISTORY:
# *************************************************************

### Third-party packages ###
from pydantic import BaseModel, StrictStr


class OpenChannel(BaseModel):
  funding_txid: StrictStr


__all__: tuple[str, ...] = ("OpenChannel",)
